// AVR High-voltage Serial Fuse Reprogrammer
// Adapted from code and design by Paul Willoughby 03/20/2010
// http://www.rickety.us/2010/03/arduino-avr-high-voltage-serial-programmer/
// Fuse Calc:
// http://www.engbedded.com/fusecalc/
// HV Program Pinouts for 84/85
// https://arduinodiy.files.wordpress.com/2015/05/fusebit.jpg

#include <Wire.h>

#define RST 13 // Output to level shifter for !RESET from transistor
#define SCI 12 // Target Clock Input
#define SDO 11 // Target Data Output
#define SII 10 // Target Instruction Input
#define SDI 7 // Target Data Input
#define VCC 8 // Target VCC

#define BTN 2   //Start Button
#define SDA A4  //I2C data for oled
#define SCL A5  //I2C clock for oled
#define SPK 3   //Buzzer
#define PWM 9   //12V pwm switching dcdc control
#define FB  A0  //12V dcdc analog feedback

#define OLEDaddr 0x3C //oled I2C address
#define CMD 0         //command packet
#define DAT 0x40      //data (image) packet

//Display in normal or inverted pixels
#define NORM 0
#define INV  1

#define HFUSE 0x747C
#define LFUSE 0x646C
#define EFUSE 0x666E

// Define ATTiny series signatures
#define ATTINY13 0x9007 // L: 0x6A, H: 0xFF 8 pin
#define ATTINY24 0x910B // L: 0x62, H: 0xDF, E: 0xFF 14 pin
#define ATTINY25 0x9108 // L: 0x62, H: 0xDF, E: 0xFF 8 pin
#define ATTINY44 0x9207 // L: 0x62, H: 0xDF, E: 0xFFF 14 pin
#define ATTINY45 0x9206 // L: 0x62, H: 0xDF, E: 0xFF 8 pin
#define ATTINY84 0x930C // L: 0x62, H: 0xDF, E: 0xFFF 14 pin
#define ATTINY85 0x930B // L: 0x62, H: 0xDF, E: 0xFF 8 pin

//PID gains
const double p_gain=0.0012; //manually tuned
const double i_gain=0.15;
const double d_gain=0.0015;
const double fbgain=3.2;    //feedback attenuation, g_fb=(R1+R2)/R1
const double supply=5000;//4710; //ADC reference supply voltage in mV (measured)
const double mult=supply*fbgain/1024; //precalculate measurement gain multiplier to speed up interrupt calculation

double reference=12000; //desired regulated output voltage in mV
double measurement=0;   //measured feedback voltage in mV
double error=0;         //PID calculation variables
double last_error=0;
double integral=0;
double derivative=0;
double output=0;        //calculated output duty
const double limit=240; //output pwm max limiter

unsigned int signature=0;
int l,h,e;

const char SmallFont[] PROGMEM = {0x05,0x08,  //5x7 Small 8 bit Font
                                0x00,0x00,0x00,0x00,0x00, // ' '  , 480 bytes
                                0x00,0x00,0x5F,0x00,0x00, // !
                                0x00,0x07,0x00,0x07,0x00, // "
                                0x14,0x7F,0x14,0x7F,0x14, // #
                                0x24,0x2A,0x7F,0x2A,0x12, // $
                                0x23,0x13,0x08,0x64,0x62, // %
                                0x36,0x49,0x55,0x22,0x50, // &
                                0x00,0x05,0x03,0x00,0x00, // '
                                0x00,0x1C,0x22,0x41,0x00, // (
                                0x00,0x41,0x22,0x1C,0x00, // )
                                0x14,0x08,0x3E,0x08,0x14, // *
                                0x08,0x08,0x3E,0x08,0x08, // +
                                0x00,0x50,0x30,0x00,0x00, // ,
                                0x08,0x08,0x08,0x08,0x08, // -
                                0x00,0x60,0x60,0x00,0x00, // .
                                0x40,0x20,0x10,0x08,0x04, // /
              
                                0x3E,0x51,0x49,0x45,0x3E, //0
                                0x00,0x42,0x7F,0x40,0x00, //1
                                0x42,0x61,0x51,0x49,0x46, //2
                                0x22,0x49,0x49,0x49,0x36, //3
                                0x18,0x14,0x12,0x7F,0x10, //4
                                0x27,0x45,0x45,0x45,0x39, //5
                                0x3E,0x49,0x49,0x49,0x32, //6
                                0x01,0x01,0x71,0x09,0x07, //7
                                0x36,0x49,0x49,0x49,0x36, //8
                                0x26,0x49,0x49,0x49,0x3E, //9
                                0x00,0x36,0x36,0x00,0x00, // :
                                0x00,0x56,0x36,0x00,0x00, // ;
                                0x08,0x14,0x22,0x41,0x00, // <
                                0x14,0x14,0x14,0x14,0x14, // =
                                0x41,0x22,0x14,0x08,0x00, // >
                                0x02,0x01,0x51,0x09,0x06, // ?
             
                                0x32,0x49,0x79,0x41,0x3E, // @
                                0x7C,0x12,0x11,0x12,0x7C, //A
                                0x7F,0x49,0x49,0x49,0x36, //B
                                0x3E,0x41,0x41,0x41,0x22, //C
                                0x7F,0x41,0x41,0x41,0x3E, //D
                                0x7F,0x49,0x49,0x49,0x41, //E
                                0x7F,0x09,0x09,0x09,0x01, //F
                                0x3E,0x41,0x49,0x49,0x7A, //G
                                0x7F,0x08,0x08,0x08,0x7F, //H
                                0x00,0x41,0x7F,0x41,0x00, //I
                                0x20,0x41,0x41,0x3F,0x01, //J
                                0x7F,0x08,0x14,0x22,0x41, //K
                                0x7F,0x40,0x40,0x40,0x40, //L
                                0x7F,0x02,0x0C,0x02,0x7F, //M
                                0x7F,0x04,0x08,0x10,0x7F, //N
                                0x3E,0x41,0x41,0x41,0x3E, //O
            
                                0x7F,0x09,0x09,0x09,0x06, //P
                                0x3E,0x41,0x51,0x21,0x5E, //Q
                                0x7F,0x09,0x19,0x29,0x46, //R
                                0x26,0x49,0x49,0x49,0x32, //S
                                0x01,0x01,0x7F,0x01,0x01, //T
                                0x3F,0x40,0x40,0x40,0x3F, //U
                                0x1F,0x20,0x40,0x20,0x1F, //V
                                0x3F,0x40,0x38,0x40,0x3F, //W
                                0x63,0x14,0x08,0x14,0x63, //X
                                0x07,0x08,0x70,0x08,0x07, //Y
                                0x61,0x51,0x49,0x45,0x43, //Z
                                0x00,0x7F,0x41,0x41,0x00, // [
                                0x15,0x16,0x7C,0x16,0x15, // Yen
                                0x00,0x41,0x41,0x7F,0x00, // ]
                                0x04,0x02,0x01,0x02,0x04, // ^
                                0x40,0x40,0x40,0x40,0x40, // _
            
                                0x00,0x01,0x02,0x04,0x00, // `
                                0x20,0x54,0x54,0x54,0x78, //a
                                0x7F,0x48,0x44,0x44,0x38, //b
                                0x38,0x44,0x44,0x44,0x20, //c
                                0x38,0x44,0x44,0x48,0x7F, //d
                                0x38,0x54,0x54,0x54,0x18, //e
                                0x08,0x7E,0x09,0x01,0x02, //f
                                0x0C,0x52,0x52,0x52,0x3E, //g
                                0x7F,0x08,0x04,0x04,0x78, //h
                                0x00,0x44,0x7D,0x40,0x00, //i
                                0x20,0x40,0x44,0x3D,0x00, //j
                                0x7F,0x10,0x28,0x44,0x00, //k
                                0x00,0x41,0x7F,0x40,0x00, //l
                                0x7C,0x04,0x18,0x04,0x78, //m
                                0x7C,0x08,0x04,0x04,0x78, //n
                                0x38,0x44,0x44,0x44,0x38, //o
           
                                0x7C,0x14,0x14,0x14,0x08, //p
                                0x08,0x14,0x14,0x18,0x7C, //q
                                0x7C,0x08,0x04,0x04,0x08, //r
                                0x48,0x54,0x54,0x54,0x24, //s
                                0x04,0x3F,0x44,0x40,0x20, //t
                                0x3C,0x40,0x40,0x20,0x7C, //u
                                0x1C,0x20,0x40,0x20,0x1C, //v
                                0x3C,0x40,0x38,0x40,0x3C, //w
                                0x44,0x28,0x10,0x28,0x44, //x
                                0x0C,0x50,0x50,0x50,0x3C, //y
                                0x44,0x64,0x54,0x4C,0x44, //z
                                0x00,0x08,0x36,0x41,0x00, // {
                                0x00,0x00,0x7F,0x00,0x00, // |
                                0x00,0x41,0x36,0x08,0x00, // }
                                0x08,0x08,0x2A,0x1C,0x08, // ->
                                0x08,0x1C,0x2A,0x08,0x08};// <-

const uint8_t logo[] PROGMEM = {32,32,  //32x32 , 128 bytes sjm4306 logo
                                0x00, 0x80, 0x80, 0xE0, 0xE0, 0x18, 0x18, 0x1E, 0x1E, 0x18, 0x18, 0x1E, 0x1E,
                                0x18, 0x18, 0x1E, 0x1E, 0x18, 0x18, 0x9E, 0x9E, 0x98, 0x98, 0x9E, 0x9E, 0x18,
                                0x18, 0xE0, 0xE0, 0x80, 0x80, 0x00, 0x00, 0x99, 0x99, 0xFF, 0xFF, 0x00, 0x00,
                                0x00, 0x00, 0x00, 0x00, 0xFE, 0xFE, 0x86, 0x86, 0x86, 0x86, 0x86, 0x86, 0x9F,
                                0x9F, 0x19, 0x19, 0x1F, 0x1F, 0x00, 0x00, 0xFF, 0xFF, 0x99, 0x99, 0x00, 0x00,
                                0x99, 0x99, 0xFF, 0xFF, 0x00, 0x00, 0xF8, 0xF8, 0x98, 0x98, 0xF9, 0xF9, 0x61,
                                0x61, 0x61, 0x61, 0x61, 0x61, 0x7F, 0x7F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                0xFF, 0xFF, 0x99, 0x99, 0x00, 0x00, 0x01, 0x01, 0x07, 0x07, 0x18, 0x18, 0x79,
                                0x79, 0x19, 0x19, 0x79, 0x79, 0x18, 0x18, 0x78, 0x78, 0x18, 0x18, 0x78, 0x78,
                                0x18, 0x18, 0x78, 0x78, 0x18, 0x18, 0x07, 0x07, 0x01, 0x01, 0x00
                               };

void writeI2C(unsigned char devaddr, unsigned char reg, bool b, unsigned char data0, unsigned char data1){  //send 1 or 2 I2C bytes to a specified device address
  Wire.beginTransmission(devaddr); //Write
  Wire.write(reg);       //Word Address
  Wire.write(data0);     //Data
  if(b){Wire.write(data1);}
  Wire.endTransmission(); //Send bytes and stop transmitting
  //delayMicroseconds(10);
}

void brightness(unsigned char brt){ //set oled screen brightness from 0-255
  writeI2C(OLEDaddr,CMD,1,0x81,brt); //Set contrast control
}

void disp_init(void){ //power cycle oled to reset it and reinitialize it (including clearing gddram)
  //pinMode(VOL,OUTPUT);
  //digitalWrite(VOL,LOW);
  //delay(50);
  //digitalWrite(VOL,HIGH);
  //delay(50);

  Wire.begin();
  Wire.setClock(400000);  //100kHz
  
  writeI2C(OLEDaddr,CMD,0,0xAE,0);    //Set display off
  writeI2C(OLEDaddr,CMD,1,0xD5,0x80); //Set display clock divide ratio, Set display oscillator freq
  writeI2C(OLEDaddr,CMD,1,0xA8,0x1F); //Set multiplex ratio
  writeI2C(OLEDaddr,CMD,1,0xD3,0x00); //Set display offset
  writeI2C(OLEDaddr,CMD,0,0x40,0);    //Set display start line
  writeI2C(OLEDaddr,CMD,1,0x8D,0x14); //Set charge pump
  writeI2C(OLEDaddr,CMD,0,0xA1,0);    //Set segment remap
  writeI2C(OLEDaddr,CMD,0,0xC8,0);    //Set com output scan dir
  writeI2C(OLEDaddr,CMD,1,0xDA,0x02); //Set com pins hardware config
  writeI2C(OLEDaddr,CMD,1,0x81,0x8F); //Set contrast control
  writeI2C(OLEDaddr,CMD,1,0xD9,0xF1); //Set precharge period
  writeI2C(OLEDaddr,CMD,1,0xDB,0x40); //Set vcomh deselect level
  writeI2C(OLEDaddr,CMD,0,0x2E,0);    //Deactivate Scrolling
  writeI2C(OLEDaddr,CMD,0,0x10,0);    //Set Higher Column Start Address
  writeI2C(OLEDaddr,CMD,0,0x00,0);    //Set Lower Column Start Address
  writeI2C(OLEDaddr,CMD,0,0xA4,0);    //Set entire display on/off
  writeI2C(OLEDaddr,CMD,0,0xA6,0);    //Set normal/inverse display
  writeI2C(OLEDaddr,CMD,1,0x20,0x00); //Set memory addressing mode
  clr();
  writeI2C(OLEDaddr,CMD,0,0xAF,0x00); //Set display on
}

void set_xy(char x, char y){  //move X,Y cursor location to write image data (0<=X<=127, 0<=Y<=7)
  Wire.beginTransmission(OLEDaddr);
  Wire.write(CMD);
  Wire.write(0x21); //set x
  Wire.write(0x00+x);
  Wire.write(0x7F);
  Wire.endTransmission();
  delayMicroseconds(10);
  writeI2C(OLEDaddr,CMD,0,0xB0+y,0); //set y
}

void clr(void){ //clear entire screen by writing 0 to all pixels stored in gddram, and reset cursor to 0,0
  set_xy(0,0);
  for(int i=0;i<8192;i++){
    writeI2C(OLEDaddr,DAT,0,0x00,0);
  }
  set_xy(0,0);
}

void printchar(char x, char y, char c, int font, bool inv){ //writes a single character to location X,Y on screen, can specify font and inversion
  unsigned char w,h=0;
  w = pgm_read_byte_near(font);
  h = pgm_read_byte_near(font + 1);
  for(int j=0;j<h/8;j++){
    set_xy(x,y+j);
    for(int i=0;i<w;i++){
      writeI2C(OLEDaddr,0b01000000,0,pgm_read_byte_near(font+w*j+2+(w*h/8)*(c-32)+i)^0xFF*inv,0);
    }
    writeI2C(OLEDaddr,DAT,0,0xFF*inv,0);
  }
}

void writeString(char x, char y, String str, int font, bool inv){ //similar to printchar, but writes a whole string to the screen
  unsigned char len = str.length();
  unsigned char w = pgm_read_byte_near(font);
  unsigned char p=0;
  while(p<len){
    printchar(x+p*(w+1),y,str.charAt(p),font,inv);
    p++;
  }
}

void drawBMP(int x, int y, unsigned long bmp, unsigned char elem, bool inv){  //write bitmap image to location X,Y on screen, can specify array bmp and elem index of image
  unsigned char w=pgm_read_byte(bmp);
  unsigned char h=pgm_read_byte(bmp+1);
  for(int j=0;j<h/8;j++){
    set_xy(x,y+j);
    for(int i=0;i<w;i++){
      writeI2C(OLEDaddr,DAT,0,pgm_read_byte_near(bmp+2+j*w+i+w*(h/8)*elem)^(inv*0xFF),0);
    }
  }
}

void init_PWM(void){  //initialize PWM for D9/OC1A (Timer1), 31.25 kHz, 8 bit resolution
  //Dont mess with Timer0 which is used for delay timing
  TCCR1A = _BV(COM1A1) | _BV(WGM12) | _BV(WGM11);  //8 bit (set TOP below) fast pwm mode 14 on OC1A pin (OC1B pin is regular IO)
  TCCR1B = _BV(CS10) | _BV(WGM13); //prescale by 1
  OCR1A = 0;  //initialize to 0% duty
  ICR1L = 127;//255;  //set TOP of timer1 for 8 bit resolution
}

void init_ADC(void){
  ADCSRA = bit (ADPS2);     // 1/16 prescale
  ADMUX = bit (REFS0) | 0;  //A0
  ADCSRA |= bit (ADEN);     //turn ADC on
}

void init_timer2(){ //Initialize Timer 2, used to trigger boost converter compensation calculations
  TCCR2A=0b00000010; //OC1A+B pins disconnected, CTC Mode
  TCCR2B=0b00001100; //32x prescaler, CTC Mode
  OCR2A=125;//249;  //Compare value, sets interrupt frequency (2kHz interrupt)
  TIMSK2=0x02; //CMPA2 Interrupt Enable
  asm("sei"); //Enable Global Interrupts
}

ISR(TIMER2_COMPA_vect){ //interrupt on Timer 2 compare A, runs at 2kHz to calculate boost converter output
  //PORTB|=0b00100000;  //LED High
  //executes at 2kHz via interrupt, execution takes ~140uS
  //measurement=supply*fbgain*((double) analogRead(FB)/1024);  //in mV
  ADCSRA |= (1<<ADSC); //start conversion
  while((ADCSRA>>ADSC)&0b1){;}  //wait till conversion done
  measurement=mult*ADC; //grab 10 bit measurement 14.6,13.8
  if(measurement>13000){
    OCR1A=0;//digitalWrite(VCC,HIGH);
  }else{
    //digitalWrite(VCC,LOW);
  }

  //Serial.println(measurement);
  
  error=reference-measurement;
  integral+=error;  //dt*error
  derivative=error-last_error;  //()/dt
  
  if(integral>limit){integral=limit;}
  if(integral<-limit){integral=-limit;}
  
  if(derivative>limit){derivative=limit;}
  if(derivative<-limit){derivative=-limit;}

  output=p_gain*error+i_gain*integral+d_gain*derivative;

  if(output>240){output=240;} //240
  if(output<0){output=0;}

  //if(abs(error)<500){PORTB|=0b00100000;}else{PORTB&=0b11011111;}

  OCR1A=output;
  last_error=error;
  //PORTB&=0b11011111;  //LED Low
}

void splashscreen(void){
  if(1){
    drawBMP(0,0,logo,0,NORM);
    writeString(33,0,"AttinyHVProg",SmallFont,NORM);
    writeString(33,1,"v1.0",SmallFont,NORM);
    writeString(33,2,"sjm4306 2022",SmallFont,NORM);
    delay(2000);
  }
  clr();
}

void beep(int n){
  for(int j=0;j<n;j++){
    for(int i=0;i<100;i++){
      digitalWrite(SPK,HIGH);
      delayMicroseconds(370);
      digitalWrite(SPK,LOW);
      delayMicroseconds(370);
    }
    delay(250);
  }
}

void setup(){
  pinMode(VCC,OUTPUT);
  pinMode(RST,OUTPUT);
  pinMode(SDI,OUTPUT);
  pinMode(SII,OUTPUT);
  pinMode(SCI,OUTPUT);
  pinMode(SDO,OUTPUT); // Configured as input when in programming mode
  digitalWrite(RST,HIGH); // Level shifter is inverting, this shuts off 12V

  pinMode(4,OUTPUT);  //necessary for clk dev board to release i2c bus
  digitalWrite(4,HIGH);
  disp_init();
  brightness(255);
  splashscreen();
  
  Serial.begin(115200);
  Serial.println("Code is modified by Rik. Visit riktronics.wordpress.com and electronics-lab.com for more projects");
  Serial.println("-------------------------------------------------------------------------------------------------");
  Serial.println("Enter any character to start process..");
  writeString(0,0,"Press Button to",SmallFont,NORM);
  writeString(0,1,"start process..",SmallFont,NORM);

  pinMode(BTN,INPUT_PULLUP);
  pinMode(SDA,INPUT_PULLUP);
  pinMode(SCL,INPUT_PULLUP);
  pinMode(SPK,OUTPUT);
  pinMode(PWM,OUTPUT);
  pinMode(FB,INPUT);
  init_PWM();
  init_ADC();
  init_timer2();
}

void loop(){
  //while(1){digitalWrite(VCC,HIGH);delay(1000);digitalWrite(VCC,LOW);delay(1000);}
  //while(1){digitalWrite(VCC,!digitalRead(BTN));}
  if(Serial.available()>0||!digitalRead(BTN)){
    Serial.read();
    clr();
    beep(2);
    writeString(0,0,"Start!",SmallFont,NORM);
    delay(2000);
    pinMode(SDO,OUTPUT); // Set SDO to output
    digitalWrite(SDI,LOW);
    digitalWrite(SII,LOW);
    digitalWrite(SDO,LOW);
    digitalWrite(RST,HIGH); // 12v Off
    digitalWrite(VCC,HIGH); // Vcc On
    delayMicroseconds(20);
    digitalWrite(RST,LOW); // 12v On
    delayMicroseconds(10);
    pinMode(SDO,INPUT); // Set SDO to input
    delayMicroseconds(300);
    unsigned int sig=readSignature();
    signature=sig;
    Serial.println("Reading signature from connected ATtiny......");
    Serial.println("Reading complete..");
    Serial.print("Signature is: ");
    Serial.println(sig,HEX);
    clr();
    writeString(0,0,"Reading Signature...",SmallFont,NORM);
    delay(1000);
    readFuses();
    delay(1000);
    writeString(0,1,"             ",SmallFont,NORM);
    if(sig==ATTINY13){
      Serial.println("The ATtiny is detected as ATtiny13/ATtiny13A..");
      writeString(0,1,"ATtiny13(A)",SmallFont,NORM);
      Serial.print("LFUSE: ");
      writeFuse(LFUSE,0x6A);
      Serial.print("HFUSE: ");
      writeFuse(HFUSE,0xFF);
      Serial.println("");
    }else if(sig==ATTINY24||sig==ATTINY44||sig==ATTINY84||sig==ATTINY25||sig==ATTINY45||sig==ATTINY85){
      Serial.println("The ATtiny is detected as "); 
      if(sig==ATTINY24){Serial.println("ATTINY24..");writeString(0,1,"ATtiny24",SmallFont,NORM);}
      else if(sig==ATTINY44){Serial.println("ATTINY44..");writeString(0,1,"ATtiny44",SmallFont,NORM);}
      else if(sig==ATTINY84){Serial.println("ATTINY84..");writeString(0,1,"ATtiny84",SmallFont,NORM);}
      else if(sig==ATTINY25){Serial.println("ATTINY25..");writeString(0,1,"ATtiny25",SmallFont,NORM);}
      else if(sig==ATTINY45){Serial.println("ATTINY45..");writeString(0,1,"ATtiny45",SmallFont,NORM);}
      else if(sig==ATTINY85){Serial.println("ATTINY85..");writeString(0,1,"ATtiny85",SmallFont,NORM);}
      else{Serial.println("ERROR!!");writeString(0,1,"ERROR!!",SmallFont,NORM);}
      delay(2000);
      writeFuse(LFUSE,0x62);
      writeFuse(HFUSE,0xDF);
      writeFuse(EFUSE,0xFF);
    }
    delay(2000);
    Serial.println("Fuses will be read again to check if it's changed successfully..");
    clr();
    writeString(0,0,"Checking Fuses",SmallFont,NORM);
    readFuses();
    digitalWrite(SCI,LOW);
    digitalWrite(VCC,LOW); // Vcc Off
    digitalWrite(RST,HIGH); // 12v Off

    if(signature==ATTINY13){
    if(l==0x6A && h==0xFF){
      writeString(30,3,"Success!",SmallFont,NORM);
      beep(1);
    }else{
      writeString(30,3,"Failed!",SmallFont,NORM);
      beep(3);
    }
  }else{
    if(l==0x62 && h==0xDF && e==0xFF){
      writeString(30,3,"Success!",SmallFont,NORM);
      beep(1);
    }else{
      writeString(30,3,"Failed!",SmallFont,NORM);
      beep(3);
    }
  }

    Serial.println("");
    Serial.println("");
    Serial.println("");
    Serial.println("");
  }
}

byte shiftOut(byte val1,byte val2){
  int inBits=0;
  //Wait until SDO goes high
  while(!digitalRead(SDO)){;}
  unsigned int dout=(unsigned int)val1<<2;
  unsigned int iout=(unsigned int)val2<<2;
  for(int ii=10;ii>=0;ii--){
    digitalWrite(SDI,!!(dout&(1<<ii)));
    digitalWrite(SII,!!(iout&(1<<ii)));
    inBits<<=1;
    inBits|=digitalRead(SDO);
    digitalWrite(SCI,HIGH);
    digitalWrite(SCI,LOW);
  } 
  return inBits>>2;
}

void writeFuse(unsigned int fuse,byte val){
  Serial.println("Writing correct fuse settings to ATtiny.......");
  clr();
  writeString(0,0,"Writing Fuses",SmallFont,NORM);
  shiftOut(0x40,0x4C);
  shiftOut(val,0x2C);
  shiftOut(0x00,(byte)(fuse>>8));
  shiftOut(0x00,(byte)fuse);

  Serial.println("Writing complete..");
  writeString(0,1,"Done!",SmallFont,NORM);
}

void readFuses(){
  int t=0;
  Serial.println("Reading fuse settings from connected ATtiny.......");
  writeString(0,1,"Reading Fuses",SmallFont,NORM);
  byte val;
  shiftOut(0x04,0x4C); // LFuse
  shiftOut(0x00,0x68);
  val = shiftOut(0x00,0x6C);
  Serial.print("LFuse: ");
  Serial.print(val,HEX);
  l=val;
  writeString(0,2,"L:",SmallFont,NORM);
  if(val/16<10){t=48;}else{t=55;}
  printchar(12,2,(val/16)+t,SmallFont,NORM);
  if(val%16<10){t=48;}else{t=55;}
  printchar(18,2,(val%16)+t,SmallFont,NORM);
  shiftOut(0x04,0x4C); // HFuse
  shiftOut(0x00,0x7A);
  val=shiftOut(0x00,0x7E);
  Serial.print(", HFuse: ");
  Serial.print(val,HEX);
  h=val;
  writeString(30,2,"H:",SmallFont,NORM);
  if(val/16<10){t=48;}else{t=55;}
  printchar(42,2,(val/16)+t,SmallFont,NORM);
  if(val%16<10){t=48;}else{t=55;}
  printchar(48,2,(val%16)+t,SmallFont,NORM);
  shiftOut(0x04,0x4C); // EFuse
  shiftOut(0x00,0x6A);
  val=shiftOut(0x00,0x6E);
  Serial.print(", EFuse: ");
  Serial.println(val, HEX);
  e=val;
  writeString(60,2,"E:",SmallFont,NORM);
  if(val/16<10){t=48;}else{t=55;}
  printchar(72,2,(val/16)+t,SmallFont,NORM);
  if(val%16<10){t=48;}else{t=55;}
  printchar(78,2,(val%16)+t,SmallFont,NORM);
  Serial.println("Reading complete..");
  writeString(0,3,"Done!",SmallFont,NORM);
}

unsigned int readSignature(){
  unsigned int sig=0;
  byte val;
  for(int ii=1;ii<3;ii++){
    shiftOut(0x08,0x4C);
    shiftOut(ii,0x0C);
    shiftOut(0x00,0x68);
    val=shiftOut(0x00,0x6C);
    sig=(sig<<8)+val;
  }
  return sig;
}
